# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pymap/admin/grpc/admin.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pymap/admin/grpc/admin.proto',
  package='admin',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1cpymap/admin/grpc/admin.proto\x12\x05\x61\x64min\"|\n\rAppendRequest\x12\x0c\n\x04user\x18\x01 \x01(\t\x12\x0e\n\x06sender\x18\x02 \x01(\t\x12\x11\n\trecipient\x18\x03 \x01(\t\x12\x0f\n\x07mailbox\x18\x04 \x01(\t\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\x12\r\n\x05\x66lags\x18\x06 \x03(\t\x12\x0c\n\x04when\x18\x07 \x01(\x04\"\x8b\x01\n\x0e\x41ppendResponse\x12\x1d\n\x06result\x18\x01 \x01(\x0e\x32\r.admin.Result\x12\x12\n\nerror_type\x18\x02 \x01(\t\x12\x16\n\x0e\x65rror_response\x18\x03 \x01(\x0c\x12\x0f\n\x07mailbox\x18\x04 \x01(\t\x12\x10\n\x08validity\x18\x05 \x01(\r\x12\x0b\n\x03uid\x18\x06 \x01(\r*)\n\x06Result\x12\x0b\n\x07SUCCESS\x10\x00\x12\x12\n\x0e\x45RROR_RESPONSE\x10\x01\x32@\n\x05\x41\x64min\x12\x37\n\x06\x41ppend\x12\x14.admin.AppendRequest\x1a\x15.admin.AppendResponse\"\x00\x62\x06proto3')
)

_RESULT = _descriptor.EnumDescriptor(
  name='Result',
  full_name='admin.Result',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERROR_RESPONSE', index=1, number=1,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=307,
  serialized_end=348,
)
_sym_db.RegisterEnumDescriptor(_RESULT)

Result = enum_type_wrapper.EnumTypeWrapper(_RESULT)
SUCCESS = 0
ERROR_RESPONSE = 1



_APPENDREQUEST = _descriptor.Descriptor(
  name='AppendRequest',
  full_name='admin.AppendRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='user', full_name='admin.AppendRequest.user', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sender', full_name='admin.AppendRequest.sender', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='recipient', full_name='admin.AppendRequest.recipient', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mailbox', full_name='admin.AppendRequest.mailbox', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='admin.AppendRequest.data', index=4,
      number=5, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='flags', full_name='admin.AppendRequest.flags', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='when', full_name='admin.AppendRequest.when', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39,
  serialized_end=163,
)


_APPENDRESPONSE = _descriptor.Descriptor(
  name='AppendResponse',
  full_name='admin.AppendResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='admin.AppendResponse.result', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_type', full_name='admin.AppendResponse.error_type', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='error_response', full_name='admin.AppendResponse.error_response', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='mailbox', full_name='admin.AppendResponse.mailbox', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='validity', full_name='admin.AppendResponse.validity', index=4,
      number=5, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uid', full_name='admin.AppendResponse.uid', index=5,
      number=6, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=166,
  serialized_end=305,
)

_APPENDRESPONSE.fields_by_name['result'].enum_type = _RESULT
DESCRIPTOR.message_types_by_name['AppendRequest'] = _APPENDREQUEST
DESCRIPTOR.message_types_by_name['AppendResponse'] = _APPENDRESPONSE
DESCRIPTOR.enum_types_by_name['Result'] = _RESULT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AppendRequest = _reflection.GeneratedProtocolMessageType('AppendRequest', (_message.Message,), dict(
  DESCRIPTOR = _APPENDREQUEST,
  __module__ = 'pymap.admin.grpc.admin_pb2'
  # @@protoc_insertion_point(class_scope:admin.AppendRequest)
  ))
_sym_db.RegisterMessage(AppendRequest)

AppendResponse = _reflection.GeneratedProtocolMessageType('AppendResponse', (_message.Message,), dict(
  DESCRIPTOR = _APPENDRESPONSE,
  __module__ = 'pymap.admin.grpc.admin_pb2'
  # @@protoc_insertion_point(class_scope:admin.AppendResponse)
  ))
_sym_db.RegisterMessage(AppendResponse)



_ADMIN = _descriptor.ServiceDescriptor(
  name='Admin',
  full_name='admin.Admin',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=350,
  serialized_end=414,
  methods=[
  _descriptor.MethodDescriptor(
    name='Append',
    full_name='admin.Admin.Append',
    index=0,
    containing_service=None,
    input_type=_APPENDREQUEST,
    output_type=_APPENDRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_ADMIN)

DESCRIPTOR.services_by_name['Admin'] = _ADMIN

# @@protoc_insertion_point(module_scope)
